/*
 *	Linux NET3:	Internet Group Management Protocol  [IGMP]
 *
 *	Authors:
 *		Alan Cox <Alan.Cox@linux.org>
 *
 *	Extended to talk the BSD extended IGMP protocol of mrouted 3.6
 *
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 */

#ifndef _LINUX_IGMP_H
#define _LINUX_IGMP_H

/*
 *	IGMP protocol structures
 */

/*
 *	Header in on cable format
 */

struct igmphdr
{
	__u8 type;
	__u8 code;		/* For newer IGMP */
	__u16 csum;
	__u32 group;
};

#define IGMP_HOST_MEMBERSHIP_QUERY	0x11	/* From RFC1112 */
#define IGMP_HOST_MEMBERSHIP_REPORT	0x12	/* Ditto */
#define IGMP_DVMRP			0x13	/* DVMRP routing */
#define IGMP_PIM			0x14	/* PIM routing */
#define IGMP_TRACE			0x15
#define IGMP_HOST_NEW_MEMBERSHIP_REPORT 0x16	/* New version of 0x11 */
#define IGMP_HOST_LEAVE_MESSAGE 	0x17

#define IGMP_MTRACE_RESP		0x1e
#define IGMP_MTRACE			0x1f


/*
 *	Use the BSD names for these for compatibility
 */

#define IGMP_DELAYING_MEMBER		0x01
#define IGMP_IDLE_MEMBER		0x02
#define IGMP_LAZY_MEMBER		0x03
#define IGMP_SLEEPING_MEMBER		0x04
#define IGMP_AWAKENING_MEMBER		0x05

#define IGMP_MINLEN			8

#define IGMP_MAX_HOST_REPORT_DELAY	10	/* max delay for response to */
						/* query (in seconds)	*/

#define IGMP_TIMER_SCALE		10	/* denotes that the igmphdr->timer field */
						/* specifies time in 10th of seconds	 */

#define IGMP_AGE_THRESHOLD		400	/* If this host don't hear any IGMP V1	*/
						/* message in this period of time,	*/
						/* revert to IGMP v2 router.		*/

#define IGMP_ALL_HOSTS		htonl(0xE0000001L)
#define IGMP_ALL_ROUTER 	htonl(0xE0000002L)
#define IGMP_LOCAL_GROUP	htonl(0xE0000000L)
#define IGMP_LOCAL_GROUP_MASK	htonl(0xFFFFFF00L)

/*
 * struct for keeping the multicast list in
 */

#ifdef __KERNEL__

/* ip_mc_socklist is real list now. Speed is not argument;
   this list never used in fast path code
 */

struct ip_mc_socklist
{
	struct ip_mc_socklist	*next;
	int			count;
	struct ip_mreqn		multi;
};

struct ip_mc_list
{
	struct in_device	*interface;
	unsigned long		multiaddr;
	struct ip_mc_list	*next;
	struct timer_list	timer;
	int			users;
	atomic_t		refcnt;
	spinlock_t		lock;
	char			tm_running;
	char			reporter;
	char			unsolicit_count;
	char			loaded;
};

extern int ip_check_mc(struct in_device *dev, u32 mc_addr);
extern int igmp_rcv(struct sk_buff *);
extern int ip_mc_join_group(struct sock *sk, struct ip_mreqn *imr);
extern int ip_mc_leave_group(struct sock *sk, struct ip_mreqn *imr);
extern void ip_mc_drop_socket(struct sock *sk);
extern void ip_mr_init(void);
extern void ip_mc_init_dev(struct in_device *);
extern void ip_mc_destroy_dev(struct in_device *);
extern void ip_mc_up(struct in_device *);
extern void ip_mc_down(struct in_device *);
extern void ip_mc_dec_group(struct in_device *in_dev, u32 addr);
extern void ip_mc_inc_group(struct in_device *in_dev, u32 addr);
#endif


/**********************************************************************
 **********************************************************************/
/* IGMP Proxy ioctl defines */
#define IGMP_PROXY_BASE			    220

#define IGMP_PROXY_INIT			    (IGMP_PROXY_BASE + 0)
#define IGMP_PROXY_UNINIT	        (IGMP_PROXY_BASE + 1)

#define IGMP_PROXY_WAN_IF_ADD		(IGMP_PROXY_BASE + 2)
#define IGMP_PROXY_WAN_IF_DEL		(IGMP_PROXY_BASE + 3)

#define IGMP_PROXY_LAN_IF_ADD		(IGMP_PROXY_BASE + 4)
#define IGMP_PROXY_LAN_IF_DEL		(IGMP_PROXY_BASE + 5)

#define IGMP_PROXY_SET_CLIENT_UNSOL_INTERVAL    (IGMP_PROXY_BASE + 6)
#define IGMP_PROXY_GET_CLIENT_UNSOL_INTERVAL    (IGMP_PROXY_BASE + 7)

#define IGMP_PROXY_SET_QUERIER_COUNTERS         (IGMP_PROXY_BASE + 8)
#define IGMP_PROXY_GET_QUERIER_COUNTERS         (IGMP_PROXY_BASE + 9)

#define IGMP_PROXY_GET_CLIENT_VERSION	        (IGMP_PROXY_BASE + 10)
#define IGMP_PROXY_GET_STATUS	                (IGMP_PROXY_BASE + 11)

#define IGMP_PROXY_GET_CLIENT_GROUP_TABLE_SIZE	        (IGMP_PROXY_BASE + 12)
#define IGMP_PROXY_GET_CLIENT_GROUP_TABLE_NEXT_INDEX	(IGMP_PROXY_BASE + 13)
#define IGMP_PROXY_GET_CLIENT_GROUP_TABLE_ENTRY	        (IGMP_PROXY_BASE + 14)

#define IGMP_PROXY_GET_ROUTER_INTF_TABLE_SIZE	    (IGMP_PROXY_BASE + 15)
#define IGMP_PROXY_GET_ROUTER_INTF_TABLE_NEXT_INDEX	(IGMP_PROXY_BASE + 16)
#define IGMP_PROXY_GET_ROUTER_INTF_TABLE_ENTRY	(IGMP_PROXY_BASE + 17)

#define IGMP_PROXY_GET_CLIENT_GROUPSTATS_TABLE_SIZE	        (IGMP_PROXY_BASE + 18)
#define IGMP_PROXY_GET_CLIENT_GROUPSTATS_TABLE_NEXT_INDEX	(IGMP_PROXY_BASE + 19)
#define IGMP_PROXY_GET_CLIENT_GROUPSTATS_TABLE_ENTRY	    (IGMP_PROXY_BASE + 20)

#define IGMP_PROXY_GET_INCLUDE_SRC_TABLE_SIZE	    (IGMP_PROXY_BASE + 21)
#define IGMP_PROXY_GET_INCLUDE_SRC_TABLE_NEXT_INDEX	(IGMP_PROXY_BASE + 22)
#define IGMP_PROXY_GET_INCLUDE_SRC_TABLE_ENTRY	    (IGMP_PROXY_BASE + 23)

#define IGMP_PROXY_GET_EXCLUDE_SRC_TABLE_SIZE	    (IGMP_PROXY_BASE + 24)
#define IGMP_PROXY_GET_EXCLUDE_SRC_TABLE_NEXT_INDEX	(IGMP_PROXY_BASE + 25)
#define IGMP_PROXY_GET_EXCLUDE_SRC_TABLE_ENTRY	    (IGMP_PROXY_BASE + 26)

#define IGMP_PROXY_GET_ROUTER_HOST_TABLE_SIZE	    (IGMP_PROXY_BASE + 27)
#define IGMP_PROXY_GET_ROUTER_HOST_TABLE_NEXT_INDEX	(IGMP_PROXY_BASE + 28)
#define IGMP_PROXY_GET_ROUTER_HOST_TABLE_ENTRY	    (IGMP_PROXY_BASE + 29)

#define IGMP_PROXY_GET_CLIENT_GROUP_TABLE_MAXSIZE   (IGMP_PROXY_BASE + 30)
#define IGMP_PROXY_GET_ROUTER_HOST_TABLE_MAXSIZE    (IGMP_PROXY_BASE + 31)

#define IGMP_PROXY_SET_MCF                          (IGMP_PROXY_BASE + 32)

#define IGMP_PROXY_GET_GUI_GROUP_ENTRY              (IGMP_PROXY_BASE + 33)
#define IGMP_PROXY_GET_GUI_GROUPSTAT_ENTRY          (IGMP_PROXY_BASE + 34)

#define IGMP_PROXY_END                              (IGMP_PROXY_BASE + 34)


/* IGMP Router Timeout values and Counters */
#define IGMPRX_ROBUSTNESS			        2
#define IGMPRX_QUERY_INTERVAL		        125
#define IGMPRX_QUERY_RESPONSE_INTERVAL      10
#define IGMPRX_LAST_MEMBER_QUERY_INTERVAL	1

/* Max number of LAN Hosts per Multicast Group */
#define IGMPRX_MAX_HOSTS_PER_GROUP  8

/* IGMP Client Unsoicited Report Interval Range */
#define IGMPRX_CLIENT_UNSOL_REPORT_INTERVAL_MIN     1
#define IGMPRX_CLIENT_UNSOL_REPORT_INTERVAL_MAX     25

#define IGMP_PROXY_CFG_DISABLE 0
#define IGMP_PROXY_CFG_ENABLE  1


/**********************************************************************
 **********************************************************************/
/* IGMP Router's Querier Compatibility Mode */
typedef enum 
{
    IGMPRX_QCM_IGMPV1 = 1,
    IGMPRX_QCM_IGMPV2 = 2,
    IGMPRX_QCM_IGMPV3 = 3
} igmprx_querier_mode_t;

typedef enum 
{
    IGMP_STATUS_DISABLED = 0,
    IGMP_STATUS_ENABLED = 1,
    IGMP_STATUS_ERROR = 2
} igmp_status_t;


/**********************************************************************
 **********************************************************************/
/* 
 * igmprx_if_t must be passed as a parameter to:
 * - IGMP_PROXY_INIT - defualt if index/ip;
 * - IGMP_PROXY_LAN_IF_ADD/IGMP_PROXY_LAN_IF_DEL lan if index/ip
 */
struct igmprx_if_t 
{
    struct in_addr if_address;
    int if_index;
};

struct igmp_s 
{
    int control_socket;
};

typedef struct igmp_s igmp_t;

typedef struct igmp_wan_if_t
{
    struct igmp_wan_if_t *next;
    char if_name[IFNAMSIZ];
} igmp_wan_if_t;

typedef struct igmp_lan_if_t
{
    struct igmp_lan_if_t *next;
    char if_name[IFNAMSIZ];
} igmp_lan_if_t;

struct igmprx_querier_counters_t 
{
    int qcm;
    int fastleave;
    int robustness;
    int query_interval;
    int query_resp_interval;
    int startup_query_interval;
    int startup_query_count;
    int lastmember_query_interval;
    int lastmember_query_count;
};

struct igmprx_router_intf_req_t 
{
    /* Multicast Group */
    unsigned int mcgrp_addr;

    int tbl_index;

    /* Router Interface Table Entry */
    unsigned char if_desc[256];
    int ifindex;
    int uptime;
    int filter_mode;
    int include_num_entries;
    int exclude_num_entries;
    int host_num_entries;
};

struct igmprx_srec_req_t 
{
    /* Multicast Group */
    unsigned int mcgrp_addr;

    /* LAN ifIndex */
    int lan_ifindex;

    int tbl_index;
    int tbl_size;

    /* Include/Exclude Source Table Entry */
    int srec_addr;
};

struct igmprx_rtr_host_req_t 
{
    /* Multicast Group */
    unsigned int mcgrp_addr;

    /* LAN ifIndex */
    int lan_ifindex;

    int tbl_index;
    int tbl_size;

    /* Include/Exclude Source Table Entry */
    unsigned int host_addr;
    unsigned long host_last_report_time;
};

/* Multicast Filtering */
#define IGMPRX_MAX_MCF_GROUPS           32

#define IGMPRX_MCF_CREATE_UPDATE_MCGRP  1
#define IGMPRX_MCF_DESTROY_MCGRP        2

struct igmprx_mcfgrp_info
{
    unsigned int mcgrp_addr;
    unsigned int mcgrp_mask;
};

struct igmprx_mcfgrp_req
{
    char ifname[IFNAMSIZ];
    int grp_count;
    struct igmprx_mcfgrp_info grp[IGMPRX_MAX_MCF_GROUPS];
    int action;
};


/**********************************************************************
 **********************************************************************/
#ifdef __KERNEL__

#ifdef CONFIG_RG_IGMP_PROXY
int igmprx_upstream_filter(int ifindex, unsigned int grp_addr);
int igmprx_setsockopt(struct sock *sk, int optname, char *optval, int optlen);
int igmprx_getsockopt(struct sock *sk, int optname, char *optval, int *optlen);
int igmprx_input(struct sk_buff *skb);
void igmprx_if_up(struct in_device *in_dev);
void igmprx_if_down(struct in_device *in_dev);
void igmprx_recv(struct sk_buff *skb);
#endif /* CONFIG_RG_IGMP_PROXY */

#endif /* __KERNEL__ */



#endif
